import 'dart:async';

import '../nosql_repository.dart';

/// Defines data access methods to a specific
/// database collection.
///
/// The [TEntity] type parameter is
/// mostly a convenience parameter
/// to facilitate the usage of
/// dependency-injection containers. If
/// you are not using a DI container or
/// your DI container does not use
/// type-parameters, you can safely ignore it.
abstract class Repository<TEntity> {
  /// Inserts the map as a new entity in the database collection.
  ///
  /// This method expects a [principal] parameter, which
  /// identifies the user performing the request. Implementations
  /// may add metadata to the entity based on that principal,
  /// like for instance a key identifying a tenant and logging.
  ///
  /// If a non-null value is passed to the [permission] parameter,
  /// the implementation of this method is expected to validate
  /// if the user represented by the [principal] does have the
  /// permission before actually inserting the entity. Otherwise,
  /// an error is thrown.
  ///
  /// It is possible that the database engine makes changes
  /// to the inserted entity, like for instance adding an
  /// automatically incremented key (primary key) to the entity.
  ///
  /// Implementations of this method are required to
  /// return the updated entity.
  Future<Map<String, dynamic>> create(
    Map<String, dynamic> map,
    DbPrincipal principal, [
    String? permission,
  ]);

  /// Deletes the entity from the collection identified by the [key] parameter.
  ///
  /// This method expects a [principal] parameter, which
  /// identifies the user performing the request. Implementations
  /// are advised to confirm if the entity being deleted belongs to
  /// a tenant related to the [principal].
  ///
  /// If a non-null value is passed to the [permission] parameter,
  /// the implementation of this method is expected to validate
  /// if the user represented by the [principal] does have the
  /// permission before actually inserting the entity. Otherwise,
  /// an error is thrown.
  Future delete(
    String key,
    DbPrincipal principal, [
    String? permission,
  ]);

  /// Update an entity in the collection contained in the [map] parameter.
  ///
  /// This method expects a [principal] parameter, which
  /// identifies the user performing the request. Implementations
  /// are advised to confirm if the entity being updated belongs to
  /// a tenant related to the [principal].
  ///
  /// If a non-null value is passed to the [permission] parameter,
  /// the implementation of this method is expected to validate
  /// if the user represented by the [principal] does have the
  /// permission before actually updating the entity. Otherwise,
  /// an error is thrown.
  Future<Map<String, dynamic>> update(
    Map<String, dynamic> map,
    DbPrincipal principal, [
    String? permission,
  ]);

  /// Retrieves an entity from the collection identified by the [key] parameter.
  ///
  /// This method expects a [principal] parameter, which
  /// identifies the user performing the request. Implementations
  /// are advised to confirm if the entity being returned belongs to
  /// a tenant related to the [principal].
  ///
  /// If a non-null value is passed to the [permission] parameter,
  /// the implementation of this method is expected to validate
  /// if the user represented by the [principal] does have the
  /// permission before actually returning the entity. Otherwise,
  /// an error is thrown.
  Future<Map<String, dynamic>> get(
    String key,
    DbPrincipal principal, [
    String? permission,
  ]);

  /// Gets all the entities from the collection identified by the [key]
  /// parameter.
  /// Use this method only with collections where the number of elements in
  /// the collection (per tenant) is guaranteed to be very limited.
  ///
  /// This method expects a [principal] parameter, which
  /// identifies the user performing the request. Implementations
  /// are advised to filter the returned elements by the tenant of the
  /// [principal].
  ///
  /// If a non-null value is passed to the [permission] parameter,
  /// the implementation of this method is expected to validate
  /// if the user represented by the [principal] does have the
  /// permission before actually returning the elements. Otherwise,
  /// an error is thrown.
  Stream<Map<String, dynamic>> getAll(
    DbPrincipal principal, [
    String? permission,
  ]);

  /// Runs a query on the database and returns the resulting entities.
  ///
  /// The [query] parameter contains a query that is dependant on the
  /// database engine being used.
  ///
  /// This method is not meant to be called directly by
  /// other libraries, by [RepositorySearch] implementations.
  Stream<Map<String, dynamic>> runQuery(
    String query,
    DbPrincipal principal, {
    Map<String, dynamic>? parameters,
    String? collectionAlias,
    String? permission,
  });
}
